import json
from collections import defaultdict
from typing import Dict

import pandas as pd
import requests

from .asset import Asset
from .data_classes import AssetClass


class History:
    def __init__(
        self, data_url: str, headers: Dict[str, str], asset: Asset
    ) -> None:
        """
        Args:
            data_url: A string representing the URL of the data.
            headers: A dictionary containing the headers to be included in the request.
            asset: An instance of the Asset class representing the asset.

        """
        self.data_url = data_url
        self.headers = headers
        self.asset = asset

    ###########################################
    # ////// Get Stock Historical Data \\\\\\ #
    ###########################################
    def get_stock_data(
        self,
        symbol: str,
        start: str,
        end: str,
        timeframe: str = "1d",
        feed: str = "sip",
        currency: str = "USD",
        limit: int = 1000,
        sort: str = "asc",
        adjustment: str = "raw",
    ) -> pd.DataFrame:
        """
        Args:
            symbol: The stock symbol to fetch data for.
            start: The start date for historical data in the format "YYYY-MM-DD".
            end: The end date for historical data in the format "YYYY-MM-DD".
            timeframe: The timeframe for the historical data. Default is "1d".
            Possible values are "1m", "5m", "15m", "30m", "1h", "4h", "1d", "1w", or "1M".
            feed: The data feed source. Default is "sip".
            currency: The currency for historical data. Default is "USD".
            limit: The number of data points to fetch. Default is 1000.
            sort: The sort order for the data. Default is "asc".
            adjustment: The adjustment for historical data. Default is "raw".

        Returns:
            A pandas DataFrame containing the historical stock data for the given symbol and time range.

        Raises:
            ValueError: If the given timeframe is not one of the allowed values.

        """

        self.check_if_stock(symbol)

        # URL for historical stock data request
        url = f"{self.data_url}/stocks/{symbol}/bars"
        # Set timeframe
        timeframe_mapping: dict = {
            "1m": "1Min",
            "5m": "5Min",
            "15m": "15Min",
            "30m": "30Min",
            "1h": "1Hour",
            "4h": "4Hour",
            "1d": "1Day",
            "1w": "1Week",
            "1M": "1Month",
        }

        if timeframe not in timeframe_mapping:
            raise ValueError(
                'Invalid timeframe. Must be "1m", "5m", "15m", "30m", "1h", "4h", "1d", "1w", or "1M"'
            )

        # Parameters for historical stock data request
        params = {
            "timeframe": timeframe_mapping[
                timeframe
            ],  # Timeframe for historical data, default: 1d
            "start": start,  # Start date for historical data
            "end": end,  # End date for historical data
            "currency": currency,  # Currency for historical data, default: USD
            "limit": limit,  # Limit number of data points, default: 1000
            "adjustment": adjustment,  # Adjustment for historical data, default: raw
            "feed": feed,  # Data feed source, default: iex
            "sort": sort,  # Sort order, default: asc
        }
        symbol_data = self.get_historical_data(symbol, url, params)
        bar_data_df = self.preprocess_data(symbol_data, symbol)
        return bar_data_df

    ###########################################
    # /////// Check if Asset is Stock \\\\\\\ #
    ###########################################
    def check_if_stock(self, symbol: str) -> AssetClass:
        """Check if asset is stock
        Args:
            symbol: The symbol of the asset to be checked.

        Returns:
            AssetClass: The asset information for the given symbol.

        Raises:
            ValueError: If there is an error getting the asset information or if the asset is not a stock.
        """
        # Get asset information for the symbol
        try:
            asset = self.asset.get(symbol)
        # Raise exception if failed to get asset information
        except Exception as e:
            raise ValueError(e)
        else:
            # Check if asset is a stock
            if asset.asset_class != "us_equity":
                # Raise exception if asset is not a stock
                raise ValueError(f"{symbol} is not a stock.")
        return asset

    ###########################################
    # /////////// PreProcess Data \\\\\\\\\\\ #
    ###########################################
    @staticmethod
    def preprocess_data(
        symbol_data: list[defaultdict], symbol: str
    ) -> pd.DataFrame:
        """Prepross data
        Preprocesses the given symbol data by converting it to a pandas DataFrame and performing various
        data transformations.

        Args:
            symbol_data: A list of defaultdict objects representing the JSON response data.
            symbol: A string representing the symbol or ticker for the stock data.

        Returns:
            A pandas DataFrame containing the preprocessed historical stock data.

        """
        # Convert JSON response to dictionary
        bar_data_df = pd.DataFrame(symbol_data)

        # Add symbol column to DataFrame
        bar_data_df.insert(0, "symbol", symbol)
        # Reformat date column
        bar_data_df["t"] = pd.to_datetime(
            bar_data_df["t"].replace("[A-Za-z]", " ", regex=True)
        )
        # Rename columns for consistency
        bar_data_df.rename(
            columns={
                "t": "date",
                "o": "open",
                "h": "high",
                "l": "low",
                "c": "close",
                "v": "volume",
                "n": "trade_count",
                "vw": "vwap",
            },
            inplace=True,
        )
        # Convert columns to appropriate data types
        bar_data_df = bar_data_df.astype(
            {
                "open": "float",
                "high": "float",
                "low": "float",
                "close": "float",
                "symbol": "str",
                "date": "datetime64[ns]",
                "vwap": "float",
                "trade_count": "int",
                "volume": "int",
            }
        )
        # Return historical stock data as a DataFrame
        return bar_data_df

    ###########################################
    # ///////// Get Historical Data \\\\\\\\\ #
    ###########################################
    def get_historical_data(
        self, symbol: str, url: str, params: dict
    ) -> list[defaultdict]:
        """Get historical stock data
        Args:
            symbol (str): The symbol of the stock or asset for which historical data is being fetched.
            url (str): The URL of the API endpoint from where the historical data is being fetched.
            params (dict): Additional parameters to be passed along with the API request.

        Returns:
            list: A list of dictionaries containing the historical data for the specified symbol.

        Raises:
            Exception: If the API response status code is not 200, an exception is raised with the error message.


        This function fetches historical data for a given symbol by making requests to the specified API endpoint.
        It uses a page token to paginate through the response data and retrieves all available historical bars.

        Example usage:
            symbol = "AAPL"
            url = "https://api.example.com/historical"
            params = {"start_date": "2021-01-01", "end_date": "2021-01-31"}

            historical_data = get_historical_data(symbol, url, params)
        """
        page_token = None
        symbols_data = defaultdict(list)
        while True:
            params["page_token"] = page_token
            response = requests.get(url, headers=self.headers, params=params)
            if response.status_code != 200:
                raise Exception(json.loads(response.text)["message"])
            res_data = response.json()
            symbols_data[symbol].extend(res_data.get("bars", []))
            page_token = res_data.get("next_page_token", "")
            if not page_token:
                break
        return symbols_data[symbol]
